/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.beacon;

import java.io.IOException;
import java.net.BindException;
import java.net.DatagramSocket;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.PortUnreachableException;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.client.beacon.internal.RepeaterClient;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.exception.ExceptionSafeErrorHandler;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSink;
import com.aquenos.epics.jackie.common.io.ByteBufferByteSource;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessBeaconMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessRepeaterConfirmMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersionUDPMessage;
import com.aquenos.epics.jackie.common.util.DatagramChannelUtil;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * <p>
 * Channel Access repeater.
 * </p>
 * 
 * <p>
 * The repeater listens on a UDP socket, waiting for registration requests from
 * clients and CA_PROTO_RSRV_IS_UP messages from servers. These messages are
 * relayed to the registered clients. This is needed because on some systems
 * only one UDP socket can be bound to the same port and on most other systems,
 * only broadcasts are sent to multiple UDP sockets while unicasts are only sent
 * to a single socket.
 * </p>
 * 
 * <p>
 * The repeater runs within the Java VM and thus is stopped when the VM is
 * shutdown. Therefore, starting an external repeater (the
 * <code>caRepeater</code> executable from the EPICS distribution) should always
 * be preferred over using this repeater. This repeater is only provided as a
 * last resort if no <code>caRepeater</code> executable can be found.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class Repeater {

    /**
     * Desired size of the repeater socket's receive buffer (in bytes). A larger
     * receive buffer makes it less likely that we lose packets because of the
     * buffer filling up before we can read them.
     */
    private final static int DESIRED_MIN_SOCKET_RECEIVE_BUFFER_SIZE = 2048;

    private final Object startStopLock = new Object();

    private DatagramChannel channel;
    private HashMap<Integer, RepeaterClient> clients = new HashMap<Integer, RepeaterClient>();
    private ChannelAccessMessageCodec codec = new ChannelAccessMessageCodec();
    // We use a fixed buffer for the confirm message because it always has the
    // same content.
    private ByteBuffer dummyMessageBuffer = ByteBuffer
            .allocate(ChannelAccessConstants.STANDARD_HEADER_SIZE);
    private ErrorHandler errorHandler;
    // We are only interested in the first 16 bytes of a packet (the first
    // message). Therefore, we can choose the buffer size very small.
    private ByteBuffer receiveBuffer = ByteBuffer
            .allocateDirect(ChannelAccessConstants.STANDARD_HEADER_SIZE);
    private ByteBufferByteSource receiveSource = new ByteBufferByteSource();
    private int repeaterPort;
    private Thread repeaterThread;
    private ByteBuffer sendBuffer = ByteBuffer
            .allocateDirect(ChannelAccessConstants.STANDARD_HEADER_SIZE);
    private ByteBufferByteSink sendSink = new ByteBufferByteSink();
    private boolean shouldRun;

    /**
     * Creates a repeater on the specified port, using the specified error
     * handler. The socket used by the repeater is not created and bound before
     * the {@link #start()} method is called.
     * 
     * @param errorHandler
     *            error handler used for logging errors (never <code>null</code>
     *            ). The error handler must never throw exceptions. This can be
     *            ensured by wrapping it in a {@link ExceptionSafeErrorHandler}.
     * @param repeaterPort
     *            UDP port the repeater binds to. This must be a port number
     *            between 1 and 65535.
     * @throws IllegalArgumentException
     *             if the <code>repeaterPort</code> is not a valid UDP port
     *             number.
     */
    public Repeater(ErrorHandler errorHandler, int repeaterPort) {
        if (errorHandler == null) {
            throw new NullPointerException();
        }
        if (repeaterPort < 1 || repeaterPort > 65535) {
            throw new IllegalArgumentException("Invalid repeater port: "
                    + repeaterPort);
        }
        this.errorHandler = errorHandler;
        this.repeaterPort = repeaterPort;
        // We have to initialize the dummy message buffer with the right
        // content.
        ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffer(dummyMessageBuffer);
        codec.encodeMessageFromRepeater(sink,
                new ChannelAccessVersionUDPMessage(ChannelAccessVersion.V4_0,
                        false, 0),
                ChannelAccessVersion.NEWEST_SUPPORTED_VERSION);
        // We have to call getWrittenData() to release and flip the buffer.
        sink.getWrittenData();
    }

    private void run() {
        synchronized (startStopLock) {
            try {
                // We have to use blocking I/O because the Java NIO
                // implementation (at least in the Sun JRE 1.7.0.45) has a bug
                // that an empty UDP packet does not cause a select operation
                // to complete. This means that registration requests (which use
                // empty packets) are not processed until another packet (e.g. a
                // beacon) is received. This leads to timeouts when clients try
                // to register. For the repeater, using blocking I/O is okay,
                // because there are no timed tasks. The repeater will only do
                // something when a packet is received.
                channel = DatagramChannelUtil.openChannelInet4();
                try {
                    channel.socket().bind(new InetSocketAddress(repeaterPort));
                } catch (BindException e) {
                    // If another socket is already bound to the port, another
                    // repeater must be already running. This is not an error,
                    // we simply quit.
                    // The BindException could also happen because we are not
                    // allowed to bind to the specified port, but this can only
                    // happen if the user explicitly specified a port below
                    // 1024.
                    resetState();
                    return;
                }
                DatagramSocket socket = channel.socket();
                // We want another repeater to be able to reuse the port when
                // this repeater has been stopped. Technically, using
                // setReuseAddress after binding the socket has undefined
                // behavior. However, we cannot call it before binding the
                // socket as this would allow us to create a second repeater if
                // one is already running. On most platforms, calling
                // setReuseAddress after binding the socket has the result that
                // we want, so we use it even though it violates the
                // specification.
                socket.setReuseAddress(true);
                // We try to increase the socket receive buffer size if
                // necessary. If this fails, it is a non-critical error, because
                // communication might still work. However, the buffer must be
                // large enough to at least hold a single packet.
                if (socket.getReceiveBufferSize() < DESIRED_MIN_SOCKET_RECEIVE_BUFFER_SIZE) {
                    try {
                        socket.setReceiveBufferSize(DESIRED_MIN_SOCKET_RECEIVE_BUFFER_SIZE);
                    } catch (SocketException e) {
                        // This is not a severe problem, so we ignore it.
                    }
                }
                if (socket.getReceiveBufferSize() < ChannelAccessConstants.STANDARD_HEADER_SIZE) {
                    // We at least need space for a single message. We do not
                    // catch the exception here, because we consider this a
                    // fatal error, which should stop the repeater from
                    // starting.
                    socket.setReceiveBufferSize(ChannelAccessConstants.STANDARD_HEADER_SIZE);
                }
            } catch (IOException e) {
                // If there is a problem with the repeater socket, the repeater
                // cannot work and thus it makes no sense to continue.
                errorHandler
                        .handleError(Repeater.class, e,
                                "Error while trying to initialize the repeater UDP socket.");
                resetState();
                return;
            }
        }

        // The initialization has finished, so now we start the main loop where
        // we wait for packets to arrive and act on them.
        boolean errorOnReceive = false;
        while (!Thread.currentThread().isInterrupted() && shouldRun) {
            try {
                doReceive();
            } catch (IOException e) {
                // If this is the first time that we catch an IOException, this
                // is most likely just a glitch. If we catch it the second time,
                // there must be something wrong with the channel and it is best
                // if we stop the repeater.
                if (!errorOnReceive) {
                    errorOnReceive = true;
                    errorHandler
                            .handleError(Repeater.class, e,
                                    "Repeater caught an unexpected IOException while trying to receive packets.");
                } else {
                    errorHandler
                            .handleError(
                                    Repeater.class,
                                    e,
                                    "Repeater caught a second unexpected IOException while trying to receive packets, shutting down...");
                    synchronized (startStopLock) {
                        resetState();
                    }
                    return;
                }
            } catch (Throwable e) {
                // We log any other exception and continue.
                errorHandler.handleError(Repeater.class, e,
                        "Unexpected error in repeater loop.");
            }
        }

        // Finally, we want to clean up the state (close open channels, etc.).
        // We do this while holding the startStopLock, so that a start or stop
        // operation does not interfere with this.
        synchronized (startStopLock) {
            resetState();
        }
    }

    private void doReceive() throws IOException {
        // This operation blocks until there is a packet available.
        SocketAddress senderSocketAddress = channel.receive(receiveBuffer);
        Pair<Inet4Address, Integer> senderAddressAndPort = extractInet4AddressAndPort(senderSocketAddress);
        if (senderAddressAndPort == null) {
            // We ignore packets that come from an address that we cannot use.
            receiveBuffer.clear();
            return;
        }
        // We have to prepare the receive buffer for reading.
        receiveBuffer.flip();
        try {
            int remaining = receiveBuffer.remaining();
            if (remaining == 0) {
                // An empty packet is a registration request.
                registerClient(senderAddressAndPort.getLeft(),
                        senderAddressAndPort.getRight());
            } else if (remaining == ChannelAccessConstants.STANDARD_HEADER_SIZE) {
                receiveSource.appendBuffer(receiveBuffer);
                ChannelAccessMessage message;
                try {
                    message = codec.decodeMessageToRepeater(receiveSource,
                            ChannelAccessVersion.NEWEST_SUPPORTED_VERSION,
                            senderAddressAndPort.getLeft());
                } catch (Exception e) {
                    // Unexpected or malformed message, log and continue.
                    errorHandler
                            .handleError(Repeater.class, null,
                                    "Repeater received an unexpected or malformed message.");
                    return;
                }
                switch (message.getCommand()) {
                case REPEATER_REGISTER:
                    // Register new client.
                    registerClient(senderAddressAndPort.getLeft(),
                            senderAddressAndPort.getRight());
                    break;
                case CA_PROTO_RSRV_IS_UP:
                    // Relay the message to all clients. The repeater in the
                    // C library relays the whole packet, not just the
                    // CA_PROTO_RSRV_IS_UP message, but this does not look
                    // like a good idea.
                    relayBeaconMessage((ChannelAccessBeaconMessage) message);
                    break;
                default:
                    // Unexpected message, log and continue.
                    errorHandler.handleError(Repeater.class, null,
                            "Repeater received an unexpected message of type "
                                    + message.getCommand() + ".");
                    return;
                }
            } else {
                // This cannot be a valid message, thus we log and
                // ignore it.
                errorHandler.handleError(Repeater.class, null,
                        "Repeater received an unexpected packet of "
                                + remaining + " bytes.");
            }
        } finally {
            // Make sure that the receive buffer is always ready for
            // reuse, even if there has been an error.
            // If there are bytes left in the receive source, we have to
            // skip them in order to release the buffer.
            receiveSource.skip(receiveSource.remaining());
            // We also have to clear the buffer.
            receiveBuffer.clear();
        }
    }

    private void registerClient(Inet4Address address, int port) {
        // The client must connect from a local address. This can be the
        // loopback address or any other address that belongs to this host.
        if (!address.equals(Inet4AddressUtil.LOOPBACK_ADDRESS)
                && !canBind(address)) {
            return;
        }
        InetSocketAddress socketAddress = new InetSocketAddress(address, port);
        RepeaterClient newClient = clients.get(port);
        if (newClient != null) {
            // A client with the same port is already registered. As we only
            // allow connections from local addresses, this must be the same
            // client.
            try {
                sendConfirmMessage(newClient);
                return;
            } catch (IOException e) {
                // There might be something wrong with the channel. We close the
                // channel and process the client like it was a new one.
                newClient.destroy();
                clients.remove(port);
            }
        }
        try {
            newClient = new RepeaterClient(socketAddress, errorHandler);
            sendConfirmMessage(newClient);
        } catch (IOException e) {
            errorHandler.handleError(Repeater.class, e,
                    "I/O error while trying to create a new repeater client.");
            return;
        }
        // We do not want to accumulate defunct clients. If there are no beacon
        // messages, we do not check clients periodically. Therefore, we send a
        // packet each time a new client connects. If this is the first packet,
        // we will not get an error, even if the client does not exist any
        // longer. The next send, however, will cause an error and we can remove
        // the client.
        for (Iterator<RepeaterClient> oldClientIterator = clients.values()
                .iterator(); oldClientIterator.hasNext();) {
            RepeaterClient oldClient = oldClientIterator.next();
            try {
                oldClient.sendPacket(dummyMessageBuffer);
            } catch (PortUnreachableException e) {
                // A PortUnreachableException has to be expected because the
                // client might have closed its socket.
                if (!oldClient.verify()) {
                    oldClient.destroy();
                    oldClientIterator.remove();
                }
            } catch (IOException e) {
                // A different exception is unexpected. We log it but otherwise
                // treat it the same way as a PortUnreachableException. This is
                // the same thing that the C library does.
                errorHandler
                        .handleError(
                                Repeater.class,
                                e,
                                "Repeater caught an unexpected error while trying to send a packet to a client.");
                if (!oldClient.verify()) {
                    oldClient.destroy();
                    oldClientIterator.remove();
                }
            }
        }
        // We insert the new client after checking the other clients because we
        // just send a confirm message to the new client and thus do not want to
        // send a dummy message.
        clients.put(port, newClient);
    }

    private void sendConfirmMessage(RepeaterClient client) throws IOException {
        try {
            sendSink.addBuffer(sendBuffer);
            codec.encodeMessageFromRepeater(
                    sendSink,
                    new ChannelAccessRepeaterConfirmMessage(client.getAddress()),
                    ChannelAccessVersion.NEWEST_SUPPORTED_VERSION);
            sendSink.getWrittenData();
            client.sendPacket(sendBuffer);
        } finally {
            sendBuffer.clear();
        }
    }

    private boolean canBind(Inet4Address address) {
        try {
            DatagramSocket socket = new DatagramSocket();
            socket.setReuseAddress(true);
            socket.bind(new InetSocketAddress(address, 0));
            socket.close();
            return true;
        } catch (SocketException e) {
            // A socket exception could have other reasons, but must likely it
            // means that the address is not a local address.
            return false;
        }
    }

    private void relayBeaconMessage(ChannelAccessBeaconMessage message) {
        sendSink.addBuffer(sendBuffer);
        codec.encodeMessageFromRepeater(sendSink, message,
                ChannelAccessVersion.NEWEST_SUPPORTED_VERSION);
        // We do not need the return value because we know that there is only a
        // single buffer backing the byte sink. This buffer (sendBuffer) is
        // flipped by getWrittenData(), so we can use it directly.
        sendSink.getWrittenData();
        for (Iterator<RepeaterClient> clientIterator = clients.values()
                .iterator(); clientIterator.hasNext();) {
            RepeaterClient client = clientIterator.next();
            try {
                client.sendPacket(sendBuffer);
            } catch (PortUnreachableException e) {
                // A PortUnreachableException has to be expected because the
                // client might have closed its socket.
                if (!client.verify()) {
                    client.destroy();
                    clientIterator.remove();
                }
            } catch (IOException e) {
                // A different exception is unexpected. We log it but otherwise
                // treat it the same way as a PortUnreachableException. This is
                // the same thing that the C library does.
                errorHandler
                        .handleError(
                                Repeater.class,
                                e,
                                "Repeater caught an unexpected error while trying to send a packet to a client.");
                if (!client.verify()) {
                    client.destroy();
                    clientIterator.remove();
                }
            }
        }
        sendBuffer.clear();
    }

    private Pair<Inet4Address, Integer> extractInet4AddressAndPort(
            SocketAddress socketAddress) {
        if (!(socketAddress instanceof InetSocketAddress)) {
            return null;
        }
        InetSocketAddress inetSocketAddress = (InetSocketAddress) socketAddress;
        InetAddress inetAddress = inetSocketAddress.getAddress();
        if (!(inetAddress instanceof Inet4Address)) {
            return null;
        }
        return Pair.of((Inet4Address) inetAddress, inetSocketAddress.getPort());
    }

    private void resetState() {
        // We close the channel.
        if (channel != null) {
            try {
                channel.close();
            } catch (IOException e) {
                // We ignore any error that might occur while trying to close a
                // channel.
            }
            channel = null;
        }
        // We also close all channels that have been opened for client
        // connections.
        for (RepeaterClient client : clients.values()) {
            client.destroy();
        }
        clients.clear();
        // We make sure that the send and receive buffers and their sink and
        // source can be reused.
        receiveSource.skip(receiveSource.remaining());
        receiveBuffer.clear();
        sendSink.getWrittenData();
        sendBuffer.clear();
        // We set the repeater thread to null. It will stop after returning from
        // this method.
        repeaterThread = null;
    }

    /**
     * Starts the repeater. If the repeater is already started, calling this
     * method has no effect. If the repeater cannot start because of a problem
     * (e.g. it cannot create or bind the UDP socket), it will stop
     * automatically. This does not cause an exception. The thread created by
     * this method is a daemon thread, so it will not keep the VM from shutting
     * down.
     */
    public void start() {
        synchronized (startStopLock) {
            if (repeaterThread == null || !repeaterThread.isAlive()) {
                resetState();
                shouldRun = true;
                repeaterThread = new Thread(new Runnable() {

                    @Override
                    public void run() {
                        Repeater.this.run();
                    }
                }, "ca-repeater-" + repeaterPort);
                // We do not want this thread to keep the VM from stopping.
                repeaterThread.setDaemon(true);
                repeaterThread.start();
            }
        }
    }

    /**
     * Stops the repeater. If the repeater is already stopped or has not been
     * started yet, this method has no effect.
     */
    public void stop() {
        synchronized (startStopLock) {
            if (repeaterThread != null) {
                shouldRun = false;
                repeaterThread.interrupt();
            }
        }
    }

}
