/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.beacon.internal;

import java.io.IOException;
import java.net.BindException;
import java.net.DatagramSocket;
import java.net.Inet4Address;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;

import com.aquenos.epics.jackie.client.beacon.Repeater;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.util.DatagramChannelUtil;

/**
 * Client connection for the {@link Repeater}. This class is only intended for
 * use by {@link Repeater}. It is closely tied to that class and has only been
 * separated into a class of its own to keep that class smaller.
 * 
 * @author Sebastian Marsching
 */
public class RepeaterClient {

    /**
     * Desired size of the per-client sockets' send buffers (in bytes). Having a
     * larger send buffer allows us to send multiple packets at once.
     */
    private final static int DESIRED_MIN_SOCKET_SEND_BUFFER_SIZE = 2048;

    private DatagramChannel channel;
    private InetSocketAddress clientAddress;
    private ErrorHandler errorHandler;

    /**
     * Creates a new client connection. The constructor opens and connects the
     * channel.
     * 
     * @param clientAddress
     *            socket address of the client (never <code>null</code>).
     * @param errorHandler
     *            error handler to use for logging errors (never (
     *            <code>null</code>).
     * @throws IOException
     *             if the channel cannot be opened, connected, or configured.
     */
    public RepeaterClient(InetSocketAddress clientAddress,
            ErrorHandler errorHandler) throws IOException {
        assert (clientAddress != null);
        assert (errorHandler != null);
        this.clientAddress = clientAddress;
        this.errorHandler = errorHandler;
        this.channel = DatagramChannelUtil.openChannelInet4();
        boolean successful = false;
        try {
            // We connect the channel so that we receive an error if the client
            // does not listen any longer.
            this.channel.connect(clientAddress);
            DatagramSocket socket = this.channel.socket();
            // We try to increase the socket send buffer size if necessary. If
            // this fails, it is a non-critical error, because communication
            // might still work. However, the buffer must be large enough to at
            // least hold a single packet.
            if (socket.getSendBufferSize() < DESIRED_MIN_SOCKET_SEND_BUFFER_SIZE) {
                try {
                    socket.setSendBufferSize(DESIRED_MIN_SOCKET_SEND_BUFFER_SIZE);
                } catch (SocketException e) {
                    // This is not a severe problem, so we ignore it.
                }
            }
            if (socket.getSendBufferSize() < ChannelAccessConstants.STANDARD_HEADER_SIZE) {
                // We at least need space for a single message. We do not
                // catch the exception here, because we consider this a
                // fatal error, which should stop the repeater from connecting
                // to the client.
                socket.setSendBufferSize(ChannelAccessConstants.STANDARD_HEADER_SIZE);
            }
            successful = true;
        } finally {
            if (!successful) {
                try {
                    this.channel.close();
                } catch (IOException e) {
                    // We ignore anyS IOException that is triggered by closing
                    // the channel.
                }
            }
        }
    }

    /**
     * Sends a packet to the client. The packet must have a size of exactly 16
     * bytes. If the packet cannot be sent immediately it is saved to be sent
     * later. If there is already a packet queued to be sent later and this
     * packet cannot be sent now, the new packet is discarded.
     * 
     * @param packet
     *            byte buffer holding the packet (never <code>null</code>). The
     *            byte buffer is not modified (even its position is not
     *            touched).
     * @throws IOException
     *             if the send operation fails or a sent operation that was done
     *             in the background since the last call of this method failed.
     */
    public void sendPacket(ByteBuffer packet) throws IOException {
        assert (packet != null);
        // We do not want to touch the buffers position.
        packet = packet.duplicate();
        // This call will block until the data has been sent.
        channel.send(packet, clientAddress);
    }

    /**
     * Tells whether the client's socket is still bound. This is checked by
     * trying to bind a socket to the same address. If the client's socket is
     * still bound, this must result in an error.
     * 
     * @return <code>true</code> if the client is still running (its socket is
     *         still bound), <code>false</code> if it is possible to bind a
     *         socket to the client's address.
     */
    public boolean verify() {
        try {
            DatagramSocket socket = new DatagramSocket(clientAddress);
            // We want a client to be able to reuse the port after we have
            // checked it. Technically, using setReuseAddress after binding the
            // socket has undefined behavior. However, we cannot call it before
            // binding the socket as this would allow us to create the second
            // one even if the client is still running. On most platforms,
            // calling setReuseAddress after binding the socket has the result
            // that we want, so we use it even though it violates the
            // specification.
            socket.setReuseAddress(true);
            socket.close();
            return false;
        } catch (BindException e) {
            // This is not an error, it just means that the port is still in
            // use. The BindException could also indicate a problem with the
            // socket address (e.g. a privileged port to which we cannot bind),
            // but this is very unlikely. In the worst case, we will keep a
            // client in our list that does not exist any longer.
            return true;
        } catch (SocketException e) {
            // We do not expect this kind of error. However, we treat it like
            // the client is still running because we cannot know better.
            errorHandler
                    .handleError(RepeaterClient.class, e,
                            "Unexpected exception while trying to test repeater client socket.");
            return true;
        }
    }

    /**
     * Returns the IP address of the interface that is used by the client.
     * 
     * @return client IP address (never <code>null</code>).
     */
    public Inet4Address getAddress() {
        // We can safely cast to an Inet4Address here, because we know that the
        // socket address was originally created from such an address.
        return (Inet4Address) clientAddress.getAddress();
    }

    /**
     * Destroys this client connection. This closes the channel.
     */
    public void destroy() {
        try {
            channel.close();
        } catch (IOException e) {
            // We ignore any exception that might be thrown when closing the
            // channel.
        }
    }

}
