/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.exception;

import com.aquenos.epics.jackie.common.protocol.ChannelAccessErrorMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;

/**
 * <p>
 * Exception indicating a Channel Access error. This exception is caused by an
 * error message that was received over the network.
 * </p>
 * 
 * <p>
 * This exception always wraps a {@link ChannelAccessStatus} that gives more
 * information about the cause of the exception and can be retrieved with
 * {@link #getStatus()} method. It might also store an extended message that was
 * received over the network and can be retrieved with the
 * {@link #getExtendedMessage()} method.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessException extends JackieRuntimeException {

    private static final long serialVersionUID = -8650712753707092991L;

    private ChannelAccessStatus status;
    private String extendedMessage;

    /**
     * <p>
     * Constructs a new exception wrapping the provided status code. The
     * optional extended message is set to <code>null</code>.
     * </p>
     * 
     * <p>
     * The {@link #getStatus()} method can be used to retrieve the status that
     * was provided when creating the exception.
     * </p>
     * 
     * @param status
     *            Channel Access status that caused the exception.
     */
    public ChannelAccessException(ChannelAccessStatus status) {
        this(status, null);
    }

    /**
     * <p>
     * Constructs a new exception wrapping the provided status code. The
     * optional extended message might be a message that was received in a
     * {@link ChannelAccessErrorMessage}.
     * </p>
     * 
     * <p>
     * The {@link #getStatus()} and {@link #getExtendedMessage()} methods can be
     * used to retrieve the status and the extended message that were provided
     * when creating the exception.
     * </p>
     * 
     * @param status
     *            Channel Access status that caused the exception.
     * @param extendedMessage
     *            text message giving more details about the cause of the
     *            problem (may be <code>null</code>).
     */
    public ChannelAccessException(ChannelAccessStatus status,
            String extendedMessage) {
        super("Channel Access status is " + status.toString()
                + (extendedMessage != null ? " (" + extendedMessage + ")" : ""));
        this.status = status;
        this.extendedMessage = extendedMessage;
    }

    /**
     * Returns the Channel Access status that caused this exception. This is the
     * status that has been received from the network.
     * 
     * @return Channel Access status that is the cause of this exception.
     */
    public ChannelAccessStatus getStatus() {
        return status;
    }

    /**
     * Returns the extended message associated with this exception. This is a
     * message that has been received over the network together with the status
     * (as part of a {@link ChannelAccessErrorMessage}). However, some error
     * messages do not provide an extended message, so that the extended message
     * may be <code>null</code>.
     * 
     * @return extended message associated with the error or <code>null</code>
     *         if no such message is available.
     */
    public String getExtendedMessage() {
        return extendedMessage;
    }

}
