/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.ByteBuffer;

/**
 * Abstract base class for classes implementing {@link ByteSource}. This class
 * implements some common methods that most classes will not want to override.
 * In particular, it maps all get methods without an offset to their counterpart
 * with an offset, so that the number of methods to implement is reduced.
 * 
 * @author Sebastian Marsching
 */
public abstract class AbstractByteSource implements ByteSource {

    /**
     * Size of a byte in bytes.
     */
    protected final static int SIZE_OF_BYTE = Byte.SIZE / 8;

    /**
     * Size of a double in bytes.
     */
    protected final static int SIZE_OF_DOUBLE = Double.SIZE / 8;

    /**
     * Size of a float in bytes.
     */
    protected final static int SIZE_OF_FLOAT = Float.SIZE / 8;

    /**
     * Size of an int in bytes.
     */
    protected final static int SIZE_OF_INT = Integer.SIZE / 8;

    /**
     * Size of a long in bytes.
     */
    protected final static int SIZE_OF_LONG = Long.SIZE / 8;

    /**
     * Size of a short in bytes.
     */
    protected final static int SIZE_OF_SHORT = Short.SIZE / 8;

    /**
     * Default constructor. This constructor is supposed to be used by classes
     * that inherit from this class.
     */
    protected AbstractByteSource() {
    }

    private void checkDestinationArraySize(int arrayLength, int offset,
            int length) {
        if (offset < 0) {
            throw new IllegalArgumentException("Offset must not be negative.");
        }
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative.");
        }
        if ((long) offset + (long) length > (long) arrayLength) {
            throw new IndexOutOfBoundsException();
        }
    }

    @Override
    public byte getByte() {
        byte value = getByteArray(1)[0];
        return value;
    }

    @Override
    public byte[] getByteArray(int length) {
        byte[] value = new byte[length];
        getByteArray(value, 0, length);
        return value;
    }

    @Override
    public double getDouble() {
        double value = getDoubleArray(1)[0];
        return value;
    }

    @Override
    public double[] getDoubleArray(int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative");
        }
        double[] value = new double[length];
        getDoubleArray(value, 0, length);
        return value;
    }

    @Override
    public AbstractByteSource getDoubleArray(double[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        byte[] bytes = getByteArray(length * SIZE_OF_DOUBLE);
        ByteBuffer.wrap(bytes).asDoubleBuffer()
                .get(destination, offset, length);
        return this;
    }

    @Override
    public float getFloat() {
        float value = getFloatArray(1)[0];
        return value;
    }

    @Override
    public float[] getFloatArray(int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative");
        }
        float[] value = new float[length];
        getFloatArray(value, 0, length);
        return value;
    }

    @Override
    public AbstractByteSource getFloatArray(float[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        byte[] bytes = getByteArray(length * SIZE_OF_FLOAT);
        ByteBuffer.wrap(bytes).asFloatBuffer().get(destination, offset, length);
        return this;
    }

    @Override
    public int getInt() {
        int value = getIntArray(1)[0];
        return value;
    }

    @Override
    public int[] getIntArray(int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative");
        }
        int[] value = new int[length];
        getIntArray(value, 0, length);
        return value;
    }

    @Override
    public AbstractByteSource getIntArray(int[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        byte[] bytes = getByteArray(length * SIZE_OF_INT);
        ByteBuffer.wrap(bytes).asIntBuffer().get(destination, offset, length);
        return this;
    }

    @Override
    public long getLong() {
        long value = getLongArray(1)[0];
        return value;
    }

    @Override
    public long[] getLongArray(int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative");
        }
        long[] value = new long[length];
        getLongArray(value, 0, length);
        return value;
    }

    @Override
    public AbstractByteSource getLongArray(long[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        byte[] bytes = getByteArray(length * SIZE_OF_LONG);
        ByteBuffer.wrap(bytes).asLongBuffer().get(destination, offset, length);
        return this;
    }

    @Override
    public short getShort() {
        short value = getShortArray(1)[0];
        return value;
    }

    @Override
    public short[] getShortArray(int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative");
        }
        short[] value = new short[length];
        getShortArray(value, 0, length);
        return value;
    }

    @Override
    public AbstractByteSource getShortArray(short[] destination, int offset,
            int length) {
        checkDestinationArraySize(destination.length, offset, length);
        byte[] bytes = getByteArray(length * SIZE_OF_SHORT);
        ByteBuffer.wrap(bytes).asShortBuffer().get(destination, offset, length);
        return this;
    }

    private static void addByteToStringBuilder(StringBuilder sb, byte b) {
        sb.append("0x");
        if (b < 16 && b >= 0) {
            sb.append('0');
        }
        sb.append(Integer.toHexString(b & 0xff));
    }

    /**
     * Returns a string representation of this byte source. The string
     * representation contains the number of bytes remaining in this source and
     * the values of the first 1024 bytes remaining in this source in a
     * hexadecimal representation.
     * 
     * @return string representation of this byte source.
     */
    @Override
    public String toString() {
        // This method is only intended for debugging purposes. Therefore, we
        // choose a rather simple (and not necessarily efficient)
        // implementation.
        // We read from the byte source in an atomic get operation and then
        // throw an exception so that now data is actually discarded.
        // We only use the first 1024 characters because typically, a user will
        // mainly be interested in the bytes at the start of the buffer.
        int remaining = remaining();
        final int length = (remaining > 1024) ? 1024 : remaining;
        final StringBuilder sb = new StringBuilder();
        sb.append(this.getClass().getName());
        sb.append("[remaining=");
        sb.append(remaining);
        sb.append(", buffer=[");
        if (length != 0) {
            try {
                atomicGet(new AtomicGetOperation<Void>() {

                    @Override
                    public Void get() {
                        for (byte b : AbstractByteSource.this
                                .getByteArray(length)) {
                            addByteToStringBuilder(sb, b);
                            sb.append(", ");
                        }
                        throw new RuntimeException(
                                "This exception should be caught by the surrounding code.");
                    }
                });
            } catch (RuntimeException e) {
                // We can ignore this exception. We just threw it to avoid
                // discarding data.
            }
            if (remaining == length) {
                // We included the complete buffer, thus we can remove the last
                // comma and space.
                sb.delete(sb.length() - 2, sb.length());
            } else {
                sb.append("...");
            }
        }
        sb.append("]]");
        return sb.toString();
    }

}
