/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>REPEATER_REGISTER</code> message. This message can
 * only be received, not sent. It is only supported for compatibility with
 * legacy clients. New clients should simply send an empty packet instead. A
 * repeater that receives this message should treat it the same way it treats an
 * empty packet.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#REPEATER_REGISTER
 */
public final class ChannelAccessRepeaterRegisterMessage extends
        ChannelAccessMessage {

    private ChannelAccessRepeaterRegisterMessage(
            ChannelAccessMessageHeader messageHeader) {
        super(ChannelAccessCommand.REPEATER_REGISTER, messageHeader);
    }

    // This message has no state, so there is no need to override equals and
    // hashCode from the super class.

    @Override
    public String toString() {
        return command.toString();
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        // This method should never be called because we already filter this
        // type of message in the ChannelAccessMessageCodec. If this method gets
        // called, something must be wrong.
        throw new UnsupportedOperationException(
                "The REPEATER_REGISTER message cannot be sent by this implementation. Send an empty packet instead.");
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        // This method should never be called because we already filter this
        // type of message in the ChannelAccessMessageCodec. If this method gets
        // called, something must be wrong.
        throw new UnsupportedOperationException(
                "The REPEATER_REGISTER message cannot be sent by this implementation. Send an empty packet instead.");
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     */
    protected static ChannelAccessRepeaterRegisterMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.REPEATER_REGISTER
                .getCommandNumber());
        try {
            byteSource.skip(messageHeader.getPayloadSize());
        } catch (BufferUnderflowException e) {
            // We ignore a buffer underflow. The repeater protocol uses a single
            // message only per packet anyway, therefore we should always be
            // aligned for the next message, even if the payload size was wrong.
        }
        return new ChannelAccessRepeaterRegisterMessage(messageHeader);
    }

}
