/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.UnsupportedProtocolVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.Inet4AddressUtil;

/**
 * Representation of a <code>CA_PROTO_SEARCH</code> message that is sent from a
 * server to a client via UDP in reply to a
 * {@link ChannelAccessSearchClientMessage}. This message carries information
 * about how to connect to the requested channel.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_SEARCH
 */
public final class ChannelAccessSearchUDPServerMessage extends
        ChannelAccessSearchMessage {

    private final static byte[] SIX_ZERO_BYTES = new byte[6];
    private final static int UNKNOWN_SERVER_ADDRESS = (int) 0xffffffff;

    private ChannelAccessVersion version;
    private int cid;
    private Inet4Address serverAddress;
    private int serverPort;

    /**
     * Creates a <code>CA_PROTO_SEARCH</code> message that can be sent from a
     * server to a client via UDP.
     * 
     * @param version
     *            Channel Access protocol version supported by this server (or
     *            the server that is serving the channel).
     * @param cid
     *            client-side channel identifier. This is the identifier that
     *            has been sent by the client with the search request and is
     *            used by the client to map the response to the right channel.
     * @param serverAddress
     *            IP address of the server serving the channel. If this is
     *            <code>null</code> or the wildcard address, the client will try
     *            to connect to the server sending this message.
     * @param serverPort
     *            TCP port of the server serving the channel. For messages using
     *            a protocol version older than Channel Access version 4.5, the
     *            port is not sent with the message and the originating port of
     *            the UDP message is used instead.
     */
    public ChannelAccessSearchUDPServerMessage(ChannelAccessVersion version,
            int cid, Inet4Address serverAddress, int serverPort) {
        super();

        if (version == null) {
            throw new NullPointerException(
                    "Channel Acccess version must not be null.");
        }

        this.version = version;
        this.cid = cid;
        this.serverAddress = serverAddress;
        this.serverPort = serverPort;
    }

    private ChannelAccessSearchUDPServerMessage(ChannelAccessVersion version,
            ChannelAccessMessageHeader messageHeader, int cid,
            Inet4Address serverAddress, int serverPort) {
        super(messageHeader);

        this.version = version;
        this.cid = cid;
        this.serverAddress = serverAddress;
        this.serverPort = serverPort;
    }

    /**
     * Returns the channel ID that is used by the client to identify the
     * channel. This integer may be zero or negative.
     * 
     * @return client-side ID for the channel.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the IP address of the server that is serving the requested
     * channel.
     * 
     * @return IP address of the server for the channel.
     */
    public Inet4Address getServerAddress() {
        return serverAddress;
    }

    /**
     * Returns the TCP port number on which the server that is serving the
     * requested channel is listening.
     * 
     * @return TCP port number.
     */
    public int getServerPort() {
        return serverPort;
    }

    /**
     * Returns the Channel Access version supported by this server. A client can
     * use this information in order to decide which features it can use when
     * connecting to the channel.
     * 
     * @return Channel Access version supported by the server.
     */
    public ChannelAccessVersion getVersion() {
        return version;
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessSearchUDPServerMessage other = (ChannelAccessSearchUDPServerMessage) obj;
        return new EqualsBuilder().append(this.version, other.version)
                .append(this.cid, other.cid)
                .append(this.serverAddress, other.serverAddress)
                .append(this.serverPort, other.serverPort).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(version).append(cid).append(serverAddress)
                .append(serverPort).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("version=");
        sb.append(version);
        sb.append(", cid=");
        sb.append(cid);
        sb.append(", serverAddress=");
        sb.append(serverAddress != null ? serverAddress.getHostAddress()
                : "null");
        sb.append(", serverPort=");
        sb.append(serverPort);
        sb.append('}');
        return sb.toString();
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "This library does not support clients with a protocol version before Channel Access version "
                            + ChannelAccessVersion.V4_4 + ".");
        }
        int serverAddress;
        short port;
        if (version.includes(ChannelAccessVersion.V4_8)
                && this.serverAddress != null
                && !this.serverAddress.isAnyLocalAddress()) {
            serverAddress = Inet4AddressUtil.addressToInt(this.serverAddress
                    .getAddress());
        } else {
            serverAddress = UNKNOWN_SERVER_ADDRESS;
        }
        port = (short) this.serverPort;
        serializeHeader(byteSink, 8, port, 0, serverAddress, cid, version,
                maxPayloadSize);
        byteSink.putShort(this.version.getMinorVersionAsShort());
        byteSink.putByteArray(SIX_ZERO_BYTES);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "This library does not support clients with a protocol version before Channel Access version "
                            + ChannelAccessVersion.V4_4 + ".");
        }
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param originatingServerAddress
     *            IP address of the server that sent this message.
     * @param originatingServerPort
     *            UDP port of the server that sent this message.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessSearchUDPServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            Inet4Address originatingServerAddress, int originatingServerPort) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_SEARCH
                .getCommandNumber());
        int payloadSize = messageHeader.getPayloadSize();
        ChannelAccessVersion version;
        if (payloadSize >= 2) {
            version = ChannelAccessVersion.forMinorVersion(byteSource
                    .getShort());
            // Consume remaining payload.
            if (payloadSize > 2) {
                byteSource.skip(payloadSize - 2);
            }
        } else {
            version = ChannelAccessVersion.V4_0;
        }
        Inet4Address serverAddress;
        int serverPort;
        if (version.includes(ChannelAccessVersion.V4_8)) {
            int serverAddressAsInt = messageHeader.getCID();
            if (serverAddressAsInt != UNKNOWN_SERVER_ADDRESS) {
                InetAddress tempAddress;
                try {
                    tempAddress = InetAddress.getByAddress(Inet4AddressUtil
                            .addressToBytes(serverAddressAsInt));
                } catch (UnknownHostException e) {
                    throw new RuntimeException(
                            "Cannot get inet address instance for address of four bytes.",
                            e);
                }
                if (!(tempAddress instanceof Inet4Address)) {
                    throw new RuntimeException(
                            "Did not get IPv4 address where it was expected. Got address of type "
                                    + tempAddress.getClass().getName()
                                    + " instead.");
                }
                serverAddress = (Inet4Address) tempAddress;
            } else {
                serverAddress = originatingServerAddress;
            }
            serverPort = messageHeader.getDataType() & 0xffff;
        } else if (version.includes(ChannelAccessVersion.V4_5)) {
            serverAddress = originatingServerAddress;
            serverPort = messageHeader.getDataType() & 0xffff;
        } else {
            serverAddress = originatingServerAddress;
            serverPort = originatingServerPort;
        }
        return new ChannelAccessSearchUDPServerMessage(version, messageHeader,
                messageHeader.getContextSpecific(), serverAddress, serverPort);
    }

}
