/*
 * Copyright 2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.channels.DatagramChannel;

/**
 * Utility methods for dealing with {@link DatagramChannel}s.
 * 
 * @author Sebastian Marsching
 */
public final class DatagramChannelUtil {

    private DatagramChannelUtil() {
    }

    private static final Method DATAGRAM_CHANNEL_OPEN_METHOD;
    private static final boolean PROTOCOL_FAMILY_AVAILABLE;
    private static final Object PROTOCOL_FAMILY_INET;

    static {
        Object protocolFamilyInet;
        Method datagramChannelOpenMethod;
        try {
            Class<?> protocolFamilyClass = Class
                    .forName("java.net.ProtocolFamily");
            Class<?> standardProtocolFamilyClass = Class
                    .forName("java.net.StandardProtocolFamily");
            protocolFamilyInet = standardProtocolFamilyClass.getField("INET")
                    .get(null);
            datagramChannelOpenMethod = DatagramChannel.class.getMethod("open",
                    protocolFamilyClass);
            // We can only use the DatagramChannel.open(...) method, if
            // StandardProtocolFamily implements ProtocolFamily (and thus we can
            // pass StandardProtocolFamily.INET to the open(...) method, if the
            // method is public and static, and if the method returns a
            // DatagramChannel.
            if (!protocolFamilyClass
                    .isAssignableFrom(standardProtocolFamilyClass)
                    || (datagramChannelOpenMethod.getModifiers() & (Modifier.PUBLIC | Modifier.STATIC)) != (Modifier.PUBLIC | Modifier.STATIC)
                    || !DatagramChannel.class
                            .isAssignableFrom(datagramChannelOpenMethod
                                    .getReturnType())) {
                protocolFamilyInet = null;
                datagramChannelOpenMethod = null;
            }
        } catch (Throwable t) {
            protocolFamilyInet = null;
            datagramChannelOpenMethod = null;
        }
        PROTOCOL_FAMILY_INET = protocolFamilyInet;
        DATAGRAM_CHANNEL_OPEN_METHOD = datagramChannelOpenMethod;
        PROTOCOL_FAMILY_AVAILABLE = PROTOCOL_FAMILY_INET != null
                && DATAGRAM_CHANNEL_OPEN_METHOD != null;
    }

    /**
     * <p>
     * Opens a new {@link DatagramChannel}. If possible, this method will call
     * <code>DatagramChannel.open(ProtocolFamily)</code>, passing
     * <code>StandardProtocolFamily.INET</code>. This will limit the socket to
     * IPv4.
     * </p>
     * 
     * <p>
     * Unfortunately, the <code>DatagramChannel.open(ProtocolFamily)</code>
     * method is only available in JDK 1.7 and newer. In order to stay
     * compatible with JDK 1.6 (both at compile-time and run-time), this method
     * actually uses reflection for calling the
     * <code>DatagramChannel.open(ProtocolFamily)</code> method. If the method
     * cannot be called (for example because the software is running on JRE 1.6
     * or reflection fails due to security restrictions), it falls back to
     * calling {@link DatagramChannel#open()}. In this case, however, the
     * returned channel will not be restricted to IPv4.
     * </p>
     * 
     * <p>
     * Any exception thrown by <code>DatagramChannel.open()</code> or
     * <code>DatagramChannel.open(ProtocolFamily)</code> is also thrown by this
     * method. Any exception caused by reflection is caught by this method and
     * {@link DatagramChannel#open()} is called directly instead.
     * </p>
     * 
     * @return datagram channel returned by <code>DatagramChannel.open()</code>
     *         or <code>DatagramChannel.open(StandardProtocolFamily.INET)</code>
     *         (the latter is preferred).
     * @throws IOException
     *             if the called <code>open</code> method throws such an
     *             exception.
     */
    public static DatagramChannel openChannelInet4() throws IOException {
        if (PROTOCOL_FAMILY_AVAILABLE) {
            try {
                return (DatagramChannel) DATAGRAM_CHANNEL_OPEN_METHOD.invoke(
                        null, PROTOCOL_FAMILY_INET);
            } catch (InvocationTargetException e) {
                // If the exception was caused by DatagramChannel.open(...)
                // throwing an exception, we want to rethrow it. If the cause is
                // not of one of the types thrown by DatagramChannel.open(...),
                // we fall-through to using the regular DatagramChannel.open()
                // method.
                Throwable cause = e.getCause();
                if (cause instanceof IOException) {
                    throw (IOException) cause;
                } else if (cause instanceof RuntimeException) {
                    throw (RuntimeException) cause;
                } else if (cause instanceof Error) {
                    throw (Error) cause;
                }
            } catch (Exception e) {
                // If the invocation fails for any reason that is caused by
                // reflection (and not by the actual method), we fall-through to
                // using the regular DatagramChannel.open() method.
            }
        }
        return DatagramChannel.open();
    }

}
