/*
 * Copyright 2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;

import java.nio.charset.Charset;
import java.util.Arrays;

import org.junit.Test;

/**
 * Tests for the {@link ChannelAccessValue} classes. Indirectly, this also tests
 * parts of the {@link ChannelAccessValueFactory}.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessValueCodecTest
 * @see ChannelAccessValueFactoryTest
 */
public class ChannelAccessValueTest {

    private static final Charset UTF8 = Charset.forName("UTF-8");

    private static void internalAssertEquals(ChannelAccessValue<?> value1,
            ChannelAccessValue<?> value2) {
        assertEquals(value1.getType(), value2.getType());
        assertEquals(value1.getValueSize(), value2.getValueSize());
    }

    private static void internalAssertEquals(ChannelAccessAlarmValue<?> value1,
            ChannelAccessAlarmValue<?> value2) {
        internalAssertEquals((ChannelAccessValue<?>) value1,
                (ChannelAccessValue<?>) value2);
        assertEquals(value1.getAlarmSeverity(), value2.getAlarmSeverity());
        assertEquals(value1.getAlarmStatus(), value2.getAlarmStatus());
    }

    private static void internalAssertEquals(
            ChannelAccessFloatingPointControlsValue<?> value1,
            ChannelAccessFloatingPointControlsValue<?> value2) {
        internalAssertEquals((ChannelAccessNumericControlsValue<?>) value1,
                (ChannelAccessNumericControlsValue<?>) value2);
        assertEquals(value1.getPrecision(), value2.getPrecision());
    }

    private static void internalAssertEquals(
            ChannelAccessNumericControlsValue<?> value1,
            ChannelAccessNumericControlsValue<?> value2) {
        internalAssertEquals((ChannelAccessAlarmValue<?>) value1,
                (ChannelAccessAlarmValue<?>) value2);
        assertEquals(value1.getCharset(), value2.getCharset());
        assertEquals(value1.getGenericLowerAlarmLimit(),
                value2.getGenericLowerAlarmLimit());
        assertEquals(value1.getGenericLowerControlLimit(),
                value2.getGenericLowerControlLimit());
        assertEquals(value1.getGenericLowerDisplayLimit(),
                value2.getGenericLowerDisplayLimit());
        assertEquals(value1.getGenericLowerWarningLimit(),
                value2.getGenericLowerWarningLimit());
        assertEquals(value1.getGenericUpperAlarmLimit(),
                value2.getGenericUpperAlarmLimit());
        assertEquals(value1.getGenericUpperControlLimit(),
                value2.getGenericUpperControlLimit());
        assertEquals(value1.getGenericUpperDisplayLimit(),
                value2.getGenericUpperDisplayLimit());
        assertEquals(value1.getGenericUpperWarningLimit(),
                value2.getGenericUpperWarningLimit());
        assertArrayEquals(value1.getRawUnits(), value2.getRawUnits());
        assertEquals(value1.getUnits(), value2.getUnits());
    }

    private static void internalAssertEquals(ChannelAccessTimeValue<?> value1,
            ChannelAccessTimeValue<?> value2) {
        internalAssertEquals((ChannelAccessAlarmValue<?>) value1,
                (ChannelAccessAlarmValue<?>) value2);
        assertEquals(value1.getTimeNanoseconds(), value2.getTimeNanoseconds());
        assertEquals(value1.getTimeSeconds(), value2.getTimeSeconds());
        assertEquals(value1.getTimeStamp(), value2.getTimeStamp());
    }

    @Test
    public void readOnlyAcknowledgeAlarm() {
        ChannelAccessAcknowledgeAlarm value = ChannelAccessValueFactory
                .createAcknowledgeAlarm(ChannelAccessAlarmSeverity.MAJOR_ALARM);
        ChannelAccessAcknowledgeAlarm readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyAlarmAcknowledgementStatus() {
        ChannelAccessAlarmAcknowledgementStatus value = ChannelAccessValueFactory
                .createAlarmAcknowledgementStatus(Arrays.asList("abc"),
                        ChannelAccessAlarmStatus.HIGH,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, true,
                        ChannelAccessAlarmSeverity.MAJOR_ALARM, UTF8);
        ChannelAccessAlarmAcknowledgementStatus readOnlyValue = value
                .asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getCharset(), readOnlyValue.getCharset());
        assertArrayEquals(value.getRawValue(), readOnlyValue.getRawValue());
        assertEquals(value.getUnacknowledgedAlarmSeverity(),
                readOnlyValue.getUnacknowledgedAlarmSeverity());
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyControlsChar() {
        ChannelAccessControlsChar value = ChannelAccessValueFactory
                .createControlsChar(new byte[] { 1, -1, 15 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, "abc",
                        (byte) 15, (byte) -23, (byte) 10, (byte) 9, (byte) 2,
                        (byte) -1, (byte) 15, (byte) -20, UTF8);
        ChannelAccessControlsChar readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyClassName() {
        ChannelAccessClassName value = ChannelAccessValueFactory
                .createClassName(Arrays.asList("abc", "123", "456"), UTF8);
        ChannelAccessClassName readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getCharset(), readOnlyValue.getCharset());
        assertArrayEquals(value.getRawValue(), readOnlyValue.getRawValue());
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyControlsDouble() {
        ChannelAccessControlsDouble value = ChannelAccessValueFactory
                .createControlsDouble(new double[] { 1.0, -1.0, 15.0 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, (short) 2,
                        "abc", 15.0, -23.0, 10.0, 9.0, 2.0, -1.0, 15.0, -20.0,
                        UTF8);
        ChannelAccessControlsDouble readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyControlsFloat() {
        ChannelAccessControlsFloat value = ChannelAccessValueFactory
                .createControlsFloat(new float[] { 1.0f, -1.0f, 15.0f },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, (short) 2,
                        "abc", 15.0f, -23.0f, 10.0f, 9.0f, 2.0f, -1.0f, 15.0f,
                        -20.0f, UTF8);
        ChannelAccessControlsFloat readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyControlsLong() {
        ChannelAccessControlsLong value = ChannelAccessValueFactory
                .createControlsLong(new int[] { 1, -1, 15 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, "abc", 15, -23,
                        10, 9, 2, -1, 15, -20, UTF8);
        ChannelAccessControlsLong readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyControlsShort() {
        ChannelAccessControlsShort value = ChannelAccessValueFactory
                .createControlsShort(new short[] { 1, -1, 15 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, "abc",
                        (short) 15, (short) -23, (short) 10, (short) 9,
                        (short) 2, (short) -1, (short) 15, (short) -20, UTF8);
        ChannelAccessControlsShort readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyGraphicsEnum() {
        ChannelAccessGraphicsEnum value = ChannelAccessValueFactory
                .createGraphicsEnum(new short[] { 1, 22, 15 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM,
                        Arrays.asList("abc", "def", "123"), UTF8);
        ChannelAccessGraphicsEnum readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getCharset(), readOnlyValue.getCharset());
        assertEquals(value.getLabels(), readOnlyValue.getLabels());
        assertArrayEquals(value.getRawLabels(), readOnlyValue.getRawLabels());
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeChar() {
        ChannelAccessTimeChar value = ChannelAccessValueFactory.createTimeChar(
                new byte[] { 1, -1, 15 }, ChannelAccessAlarmStatus.COMM,
                ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789);
        ChannelAccessTimeChar readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeDouble() {
        ChannelAccessTimeDouble value = ChannelAccessValueFactory
                .createTimeDouble(new double[] { 1.0, -1.0, 15.0 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789);
        ChannelAccessTimeDouble readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeEnum() {
        ChannelAccessTimeEnum value = ChannelAccessValueFactory.createTimeEnum(
                new short[] { 1, 3, 15 }, ChannelAccessAlarmStatus.COMM,
                ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789);
        ChannelAccessTimeEnum readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeFloat() {
        ChannelAccessTimeFloat value = ChannelAccessValueFactory
                .createTimeFloat(new float[] { 1.0f, -1.0f, 15.0f },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789);
        ChannelAccessTimeFloat readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeLong() {
        ChannelAccessTimeLong value = ChannelAccessValueFactory.createTimeLong(
                new int[] { 1, -1, 15 }, ChannelAccessAlarmStatus.COMM,
                ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789);
        ChannelAccessTimeLong readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeShort() {
        ChannelAccessTimeShort value = ChannelAccessValueFactory
                .createTimeShort(new short[] { 1, -1, 15 },
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789);
        ChannelAccessTimeShort readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

    @Test
    public void readOnlyTimeString() {
        ChannelAccessTimeString value = ChannelAccessValueFactory
                .createTimeString(Arrays.asList("abc", "321", "xyz"),
                        ChannelAccessAlarmStatus.COMM,
                        ChannelAccessAlarmSeverity.MINOR_ALARM, 12345, 6789,
                        UTF8);
        ChannelAccessTimeString readOnlyValue = value.asReadOnlyValue();
        assertEquals(value, readOnlyValue);
        assertEquals(readOnlyValue, value);
        internalAssertEquals(value, readOnlyValue);
        assertEquals(value.getCharset(), readOnlyValue.getCharset());
        assertArrayEquals(value.getRawValue(), readOnlyValue.getRawValue());
        assertEquals(value.getValue(), readOnlyValue.getValue());
    }

}
